#include "exrIO.H"
using namespace Imf;
using namespace Imath;
using namespace std;


void ReadExr(const char* inputFile, float* & inData, int & width, int & height)
{
    InputFile file(inputFile);

    Box2i dw = file.header().dataWindow();

    // these will be overwritten later
    width = dw.max.x - dw.min.x + 1;
    height = dw.max.y - dw.min.y + 1;
    int dx = dw.min.x;
    int dy = dw.min.y;

    Box2i dispw = file.header().displayWindow();
    int dispW = dispw.max.x - dispw.min.x + 1;
    int dispH = dispw.max.y - dispw.min.y + 1;

    FrameBuffer frameBuffer;

    // specify how to store values for inData
    inData = new float[dispW * dispH * 4];
    int pixSize = sizeof(float) * 4;
    int lineSize = dispW * pixSize;

    const char* stdChannelNames[4] = { "R", "G", "B", "A" };

    // read in existing channel names
    const ChannelList &channels = file.header().channels();
    int numChannels = 0;
    for (ChannelList::ConstIterator iter = channels.begin(); iter != channels.end(); ++iter) {

        string curname(iter.name());
        int index = -1;
        for (int i = 0; i < 4; i++) {
            string expectedname(stdChannelNames[i]);
            if (expectedname == curname) {
                index = i;
                break;
            }
        }
        if (index == -1) {
            fprintf(stderr, "ReadEXR warning: '%s'\n"
                "  nonstandard channel name '%s' (expected '%s')\n",
                inputFile, curname.c_str(), stdChannelNames[numChannels]);
            if (numChannels == 3) {
                // only read in an alpha channel if it is called "A"
                break;
            }
            index = numChannels;
        }

        float defaultVal = (numChannels == 3) ? 1.0 : 0.0;
        frameBuffer.insert(curname.c_str(),
            Slice(Imf::FLOAT, (char*) & inData[index],
                pixSize, lineSize, 1, 1, defaultVal));

        numChannels++;
        if (numChannels == 4) {
            break;
        }
    }

    // fill in more channels if we read less than 4 channel names
    for (int i = numChannels; i < 4; i++) {
        float defaultVal = (i == 3) ? 1.0 : 0.0;
        frameBuffer.insert(stdChannelNames[i],
            Slice(Imf::FLOAT, (char*) & inData[i],
                pixSize, lineSize, 1, 1, defaultVal));
    }


    file.setFrameBuffer(frameBuffer);
    try {
        file.readPixels(dw.min.y, dw.max.y);
    } catch (std::exception & e) {
        fprintf(stderr, "error: could not read from file \"%s\":\n"
            "  %s\n", inputFile, e.what());
        exit(1);
    }

    width = dispW;
    height = dispH;
}

void WriteExrRGB(const char* outputFile, const float* outData, int width, int height)
{
    Header header(width, height);
    header.channels().insert("R", Channel(Imf::FLOAT));
    header.channels().insert("G", Channel(Imf::FLOAT));
    header.channels().insert("B", Channel(Imf::FLOAT));

    OutputFile file(outputFile, header);

    FrameBuffer buffer;
    int pixSize = sizeof(float) * 3;
    int lineSize = pixSize * width;
    buffer.insert("R", Slice(Imf::FLOAT, (char*) & outData[0], pixSize, lineSize));
    buffer.insert("G", Slice(Imf::FLOAT, (char*) & outData[1], pixSize, lineSize));
    buffer.insert("B", Slice(Imf::FLOAT, (char*) & outData[2], pixSize, lineSize));

    file.setFrameBuffer(buffer);
    try {
        file.writePixels(height);
    } catch (std::exception & e) {
        fprintf(stderr, "error: could not write to file \"%s\":\n"
            "  %s\n", outputFile, e.what());
        exit(1);
    }
}

void WriteExr(const char* outputFile, const float* outData, int width, int height)
{
    Header header(width, height);
    header.channels().insert("R", Channel(Imf::FLOAT));
    header.channels().insert("G", Channel(Imf::FLOAT));
    header.channels().insert("B", Channel(Imf::FLOAT));
    header.channels().insert("A", Channel(Imf::FLOAT));

    OutputFile file(outputFile, header);

    FrameBuffer buffer;
    int pixSize = sizeof(float) * 4;
    int lineSize = pixSize * width;
    buffer.insert("R", Slice(Imf::FLOAT, (char*) & outData[0], pixSize, lineSize));
    buffer.insert("G", Slice(Imf::FLOAT, (char*) & outData[1], pixSize, lineSize));
    buffer.insert("B", Slice(Imf::FLOAT, (char*) & outData[2], pixSize, lineSize));
    buffer.insert("A", Slice(Imf::FLOAT, (char*) & outData[3], pixSize, lineSize));

    file.setFrameBuffer(buffer);
    try {
        file.writePixels(height);
    } catch (std::exception & e) {
        fprintf(stderr, "error: could not write to file \"%s\":\n"
            "  %s\n", outputFile, e.what());
        exit(1);
    }
}



void ReadRaw(const char* inputFile, float* & inData, int & width, int & height)
{
    FILE* infp = fopen(inputFile, "rb");
    if (infp == NULL) {
        fprintf(stderr, "error: cannot open \"%s\" for reading\n",
            inputFile);
        exit(1);
    }

    bool error;
    error = (fread(& width, sizeof(int), 1, infp) != 1);
    error = error || (fread(& height, sizeof(int), 1, infp) != 1);
    if (error) {
        fprintf(stderr, "error: cannot read header for file \"%s\"\n",
            inputFile);
        exit(1);
    }
    if (width < 1 || width > 10240 || height < 1 || height > 10240) {
        fprintf(stderr, "error: input file \"%s\" has dimensions\n"
            "  %dx%d which are out of range\n", inputFile, width, height);
        exit(1);
    }

    int numPixels = width * height;
    int numFloats = 4 * numPixels;
    inData = new float[numFloats];
    error = (fread(inData, sizeof(float) * numFloats, 1, infp) != 1);
    if (error) {
        fprintf(stderr, "error: cannot read data for file \"%s\"\n",
            inputFile);
        exit(1);
    }

    fclose(infp);
}



void WriteRaw(const char* outputFile, const float* outData, int width, int height)
{
    int numPixels = width * height;
    int numFloats = 4 * numPixels;
    bool error;

    FILE* outfp = fopen(outputFile, "wb");
    if (outfp == NULL) {
        fprintf(stderr, "error: cannot open \"%s\" for writing\n",
            outputFile);
        exit(1);
    }

    error = (fwrite(& width, sizeof(int), 1, outfp) != 1);
    error = error || (fwrite(& height, sizeof(int), 1, outfp) != 1);
    if (error) {
        fprintf(stderr, "error: cannot write header for file \"%s\"\n",
            outputFile);
        exit(1);
    }

    error = (fwrite(outData, sizeof(float) * numFloats, 1, outfp) != 1);
    if (error) {
        fprintf(stderr, "error: cannot write data for file \"%s\"\n",
            outputFile);
        exit(1);
    }
    fclose(outfp);
}



void ReadTxt(const char* inputFile, const float* inData, int width, int height)
{
    ifstream instream;
    instream.open(inputFile);
    if (!instream.good()) {
        fprintf(stderr, "Error: could not open %s\n", inputFile);
        exit(1);
    }

    /*
    ifstream instream;
    instream.open(inputFile);
    string line;
    while (getline(instream, line)) {
        istringstream linestream(line);
        string word;
        while (line >> word) {
        }
    }
    */

    int numPixels = width * height;
    int numFloats = 4 * numPixels;
    float* indata = new float[numPixels];
    for (int i = 0; i < numPixels; i++) {
        bool success = instream >> indata[i];
        if (!success) {
            fprintf(stderr, "Error: could not read in %dx%d data from %s\n",
                width, height, inputFile);
            exit(1);
        }
    }
}



void WriteTxt(const char* outputFile, const float* outData, int width, int height)
{
    FILE* outfp = fopen(outputFile, "wb");
    if (outfp == NULL) {
        fprintf(stderr, "error: cannot open \"%s\" for writing\n",
            outputFile);
        exit(1);
    }

    for (int y = 0; y < height; y++) {
        for (int x = 0; x < width; x++) {
            float greyscale = outData[(y * width + x) * 4 + 0];
            greyscale += outData[(y * width + x) * 4 + 1];
            greyscale += outData[(y * width + x) * 4 + 2];
            greyscale /= 3.0;

            fprintf(outfp, "%f ", greyscale);
        }
        fprintf(outfp, "\n");
    }
    fclose(outfp);
}



void WriteRgb(const char* outputFile, const float* outData, int width, int height)
{
    FILE* outfp = fopen(outputFile, "wb");
    if (outfp == NULL) {
        fprintf(stderr, "error: cannot open \"%s\" for writing\n",
            outputFile);
        exit(1);
    }

    float* rgbData = new float[3 * width * height];

    for (int y = 0; y < height; y++) {
        for (int x = 0; x < width; x++) {
            rgbData[(y * width + x) * 3 + 0] = outData[(y * width + x) * 4 + 0];
            rgbData[(y * width + x) * 3 + 1] = outData[(y * width + x) * 4 + 1];
            rgbData[(y * width + x) * 3 + 2] = outData[(y * width + x) * 4 + 2];
        }
    }

    bool error = (fwrite(rgbData, sizeof(float) * 3 * width * height, 1, outfp) != 1);
    if (error) {
        fprintf(stderr, "error: cannot write data for file \"%s\"\n",
            outputFile);
        exit(1);
    }
    fclose(outfp);
    delete[] rgbData;
}




void WriteRgba(const char* outputFile, const float* outData, int width, int height)
{
    FILE* outfp = fopen(outputFile, "wb");
    if (outfp == NULL) {
        fprintf(stderr, "error: cannot open \"%s\" for writing\n",
            outputFile);
        exit(1);
    }


    bool error = (fwrite(outData, sizeof(float) * 4 * width * height, 1, outfp) != 1);
    if (error) {
        fprintf(stderr, "error: cannot write data for file \"%s\"\n",
            outputFile);
        exit(1);
    }
    fclose(outfp);
}





// h (hue) should be between 0.0 and 1.0
void RainbowtoRGB( float *r, float *g, float *b, float h) {
   int i;
   float f, p, q, t;
   float s = 1.0;  // saturation
   float v = 1.0;  // value

   if( s == 0 ) {
       // achromatic (grey)
       *r = *g = *b = v;
       return;
   }

   h *= 6;
   i = (int) floor( h );
   f = h - i;              // factorial part of h
   p = v * ( 1 - s );
   q = v * ( 1 - s * f );
   t = v * ( 1 - s * ( 1 - f ) );

   switch( i ) {
       case 0:
           *r = v;
           *g = t;
           *b = p;
           break;
       case 1:
           *r = q;
           *g = v;
           *b = p;
           break;
       case 2:
           *r = p;
           *g = v;
           *b = t;
           break;
       case 3:
           *r = p;
           *g = q;
           *b = v;
           break;
       case 4:
           *r = t;
           *g = p;
           *b = v;
           break;
       default:        // case 5:
           *r = v;
           *g = p;
           *b = q;
           break;
   }
}



